class Deanonymizer:
    def deanonymize(self, text, mappings):
        # Loop through each pattern mapping and replace the anonymized values
        # back to the original ones
        for _, pattern_map in mappings.items():
            for original_value, fake_value in pattern_map.items():
                text = text.replace(fake_value, original_value)
        return text



from functools import wraps
import re
from faker import Faker
import spacy


class Anonymizer:
    def __init__(self, custom_patterns=None, keep_default=True):
        self.fake = Faker()
        self.load_spacy_model()

        # Extract valid labels from the NLP model
        self.valid_labels = set(label for label
                                in self.nlp.get_pipe("ner").labels)

        # Define default regex patterns
        person_pattern = "PERSON"
        phone_pattern = r"\+?\d{1,4}?[-.\s]?\(?\d{1,3}?\)?[-.\s]?\d{1,4}[-.\s]?\d{1,4}[-.\s]?\d{1,9}" # noqa
        email_pattern = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
        credit_card_pattern = r'\d{4}[-\s]?\d{4}[-\s]?\d{4}[-\s]?\d{4}'
        address_pattern = r'\d{1,5}\s\w+(\s\w+)*,\s\w+,\s\w+(\s\w+)*'

        # Initialize pattern functions
        self.pattern_functions = []

        # Keep default patterns if specified
        if keep_default or custom_patterns is None:
            self.pattern_functions.extend([
                self.create_anonymize_function(person_pattern,
                                               self.fake.name),
                self.create_anonymize_function(phone_pattern,
                                               self.fake.phone_number),
                self.create_anonymize_function(email_pattern,
                                               self.fake.email),
                self.create_anonymize_function(credit_card_pattern,
                                               self.fake.credit_card_number),
                self.create_anonymize_function(address_pattern,
                                               self.fake.address),
            ])

        # Add any custom patterns
        if custom_patterns:
            for pattern, func in custom_patterns.items():
                self.pattern_functions.append(
                    self.create_anonymize_function(pattern, func))

    def load_spacy_model(self):
        try:
            self.nlp = spacy.load("en_core_web_sm")
        except: # noqa
            print("Model not found. Downloading en_core_web_sm...")
            spacy.cli.download("en_core_web_sm")
            self.nlp = spacy.load("en_core_web_sm")

    def generate_unique_fake(self, original, generator_func):
        fake_value = generator_func()
        while fake_value == original:
            fake_value = generator_func()
        return fake_value

    def create_anonymize_function(self, pattern, fake_func):
        if pattern not in self.valid_labels:
            try:
                re.compile(pattern)
            except re.error:
                raise ValueError(f"Invalid pattern: {pattern}. Must be a spaCy label or a regular expression.") # noqa

        def anonymize_func(sentence, anon_sentence, mappings):
            data_map = {}
            if pattern in self.valid_labels:
                doc = self.nlp(sentence)
                for ent in doc.ents:
                    if ent.label_ == pattern:
                        fake_data = self.generate_unique_fake(ent.text,
                                                              fake_func)
                        data_map[ent.text] = fake_data
                        anon_sentence = anon_sentence.replace(ent.text,
                                                              fake_data)
            else:
                for data in re.findall(pattern, sentence):
                    fake_data = self.generate_unique_fake(data, fake_func)
                    data_map[data] = fake_data
                    anon_sentence = anon_sentence.replace(data, fake_data)
            mappings[pattern] = data_map
            return anon_sentence
        return anonymize_func

    def anonymize_data(self, sentence):
        anon_sentence = sentence
        mappings = {}
        for pattern_function in self.pattern_functions:
            anon_sentence = pattern_function(sentence, anon_sentence, mappings)
        return anon_sentence, mappings

    def anonymize(self, *args_to_anonymize):
        def inner_decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                for arg_name in args_to_anonymize:
                    if arg_name in kwargs:
                        anonymized_data, _ = self.anonymize_data(
                            kwargs[arg_name])
                        kwargs[arg_name] = anonymized_data
                return func(*args, **kwargs)
            return wrapper
        return inner_decorator



import unittest
from anonLLM.deanonymizer import Deanonymizer
from anonLLM.anonymizer import Anonymizer
import re


class TestDeanonymizer(unittest.TestCase):
    def setUp(self):
        self.deanonymizer = Deanonymizer()
        self.anonymizer = Anonymizer()

    def extract_info(self, text):
        name_pattern = r"My name is ([\w\s]+),"
        email_pattern = r"email: ([\w\.-]+@[\w\.-]+),"
        phone_pattern = r"phone: ([\+\d\s\-]+)\."

        name = re.search(name_pattern, text).group(1)
        email = re.search(email_pattern, text).group(1)
        phone = re.search(phone_pattern, text).group(1)

        return name, email, phone

    def test_deanonymize(self):
        # Test with 10 different examples that include name, email, and phone numbers # noqa
        test_examples = [
            "My name is Alice Johnson, email: alice.johnson@example.com, "
            "phone: +1 234-567-8910.",

            "My name is Bob Smith, email: bob_smith@example.com, "
            "phone: +33 1 23 45 67 89.",

            "My name is Charlie Brown, email: charlie.brown@example.com, "
            "phone: +237 6 1234 5678.",

            "My name is David Wang, email: david.wang@example.com, "
            "phone: +1 987-654-3210.",

            "My name is Eve Adams, email: eve.adams@example.com, "
            "phone: +33 9 87 65 43 21.",

            "My name is Frank Lee, email: frank.lee@example.com, "
            "phone: +237 6 8765 4321.",

            "My name is Grace Kim, email: grace.kim@example.com, "
            "phone: +1 555-444-3333.",

            "My name is Harry Johnson, email: harry.johnson@example.com, "
            "phone: +33 4 56 78 90 12.",

            "My name is Irene Williams, email: irene.williams@example.com, "
            "phone: +237 6 7890 1234.",

            "My name is John Doe, email: john.doe@example.com, "
            "phone: +1 111-222-3333."
        ]

        for example in test_examples:
            anonymized_text, mapping = self.anonymizer.anonymize_data(example)
            deanonymized_text = self.deanonymizer.deanonymize(anonymized_text,
                                                              mapping)

            name, email, phone = self.extract_info(example)

            self.assertIn(name, deanonymized_text)
            self.assertIn(email, deanonymized_text)
            self.assertIn(phone, deanonymized_text)


if __name__ == "__main__":
    unittest.main()



import unittest

from pydantic import BaseModel

from anonLLM.llm import OpenaiLanguageModel
from dotenv import load_dotenv

load_dotenv()


class TestOpenaiLanguageModel(unittest.TestCase):
    def setUp(self):
        self.llm = OpenaiLanguageModel()

    def test_generate(self):
        prompt = (
            "My name is Alice Johnson, "
            "email: alice.johnson@example.com, "
            "phone: +1 234-567-8910. "
            "Write an imaginary cover letter for me "
            "as a machine learning engineer."
        )
        response = self.llm.generate(prompt)

        self.assertIsNotNone(response)
        self.assertNotEqual("", response.strip())
        self.assertIn("Alice Johnson", response)

    def test_generate_with_model_output(self):
        class Output(BaseModel):
            person: str
            city: str

        prompt = 'Extract the requested  information from the following sentence: "Alice Johnson is visiting Rome."'
        response = self.llm.generate(prompt, output_format=Output)
        self.assertTrue(isinstance(response, Output))
        self.assertIn("Alice Johnson", response.person)
        self.assertIn("Rome", response.city)

    def test_generate_n_completions(self):
        prompt = (
            "What is the user's favorite color in the following expression?"
            "\nAlice Johnson's favorite color is blue"
        )
        responses = self.llm.generate(prompt, n_completions=2)
        self.assertEqual(len(responses), 2)
        for response in responses:
            self.assertIsNotNone(response)
            self.assertIn("blue", response.lower())

    def test_generate_n_completions_with_model_output(self):
        class Output(BaseModel):
            person: str
            city: str

        prompt = 'Extract the requested  information from the following sentence: "Alice Johnson is visiting Rome."'
        responses = self.llm.generate(prompt, output_format=Output, n_completions=2)
        self.assertEqual(len(responses), 2)
        for response in responses:
            self.assertTrue(isinstance(response, Output))
            self.assertIn("Alice Johnson", response.person)
            self.assertIn("Rome", response.city)



if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="anonLLM",
    version="0.1.8",
    author="FS Ndzomga",
    author_email="ndzomgafs@gmail.com",
    description="Anonymizes personally identifiable information for "
    "Large Language Model APIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fsndzomga/anonLLM",
    project_urls={
        "Bug Tracker": "https://github.com/fsndzomga/anonLLM/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=[
        "aiohttp>=3.8,<3.9",
        "aiosignal>=1.3,<1.4",
        "annotated-types>=0.5,<0.6",
        "async-timeout>=4.0,<4.1",
        "attrs>=23.1,<24.0",
        "blis>=0.7,<0.8",
        "catalogue>=2.0,<2.1",
        "certifi>=2023.7,<2024.0",
        "charset-normalizer>=3.2,<3.3",
        "click>=8.1,<8.2",
        "confection>=0.1,<0.2",
        "cymem>=2.0,<2.1",
        "Faker>=19.5,<20.0",
        "frozenlist>=1.4,<1.5",
        "idna>=3.4,<3.5",
        "Jinja2>=3.1,<3.2",
        "langcodes>=3.3,<3.4",
        "MarkupSafe>=2.1,<2.2",
        "multidict>=6.0,<6.1",
        "murmurhash>=1.0,<1.1",
        "numpy>=1.25,<1.26",
        "openai>=0.28,<0.29",
        "packaging>=23.1,<24.0",
        "pathy>=0.10,<0.11",
        "pip>=22.3,<22.4",
        "preshed>=3.0,<3.1",
        "pydantic>=2.3,<2.4",
        "pydantic_core>=2.6,<2.7",
        "python-dateutil>=2.8,<2.9",
        "python-dotenv>=1.0,<1.1",
        "regex>=2023.8,<2023.9",
        "requests>=2.31,<2.32",
        "setuptools>=65.5,<65.6",
        "six>=1.16,<1.17",
        "smart-open>=6.4,<6.5",
        "spacy>=3.6,<3.7",
        "spacy-legacy>=3.0,<3.1",
        "spacy-loggers>=1.0,<1.1",
        "srsly>=2.4,<2.5",
        "thinc>=8.1,<8.2",
        "tiktoken>=0.4,<0.5",
        "tqdm>=4.66,<4.67",
        "typer>=0.9,<1.0",
        "typing_extensions>=4.7,<4.8",
        "urllib3>=2.0,<2.1",
        "wasabi>=1.1,<1.2",
        "yarl>=1.9,<1.10"
    ],
    extras_require={
        "dev": ["pytest", "flake8"],
    },
)



import openai
import time
import json
from pydantic import BaseModel, ValidationError, ConfigDict
from typing import Type, Optional
import os
from anonLLM.anonymizer import Anonymizer
from anonLLM.deanonymizer import Deanonymizer


class OpenaiLanguageModel:
    def __init__(self, api_key=None, model="gpt-3.5-turbo", temperature=0.5, anonymize=True):
        self.anonymize = anonymize

        if self.anonymize:
            self.anonymizer = Anonymizer()
            self.deanonymizer = Deanonymizer()

        if api_key is None:
            api_key = os.environ.get('OPENAI_API_KEY')

        if api_key is None:
            raise ValueError("The OPENAI API KEY must be provided either as "
                             "an argument or as an environment variable named 'OPENAI_API_KEY'") # noqa

        self.api_key = api_key
        self.model = model
        self.temperature = temperature
        openai.api_key = self.api_key

    def generate(self, prompt: str, output_format: Optional[Type[BaseModel]] = None,
                 n_completions: int = 1, max_tokens: int = None):
        anonymized_prompt, mappings = (self.anonymizer.anonymize_data(prompt)
                                       if self.anonymize else (prompt, None))

        retry_delay = 0.1
        valid_responses = []

        while len(valid_responses) < n_completions:
            try:
                system_message = "You are a helpful assistant."
                if output_format:
                    system_message += f" Respond in a JSON format that contains the following keys: {self._model_structure_repr(output_format)}" # noqa

                params = {
                    "model": self.model,
                    "messages": [
                        {
                            "role": "system",
                            "content": system_message
                        },
                        {
                            "role": "user",
                            "content": anonymized_prompt
                        }
                    ],
                    "temperature": self.temperature,
                    "n": n_completions
                }

                if max_tokens is not None:
                    params["max_tokens"] = max_tokens

                response = openai.ChatCompletion.create(**params)
                choices = response["choices"]
                responses = [choice["message"]["content"]
                             for choice in choices]

                if output_format:
                    valid_responses.extend(
                        [json.loads(res) for res in responses
                         if self._is_valid_json_for_model(res, output_format)]
                    )
                else:
                    valid_responses.extend(responses)

            except openai.error.RateLimitError:
                print(f"Hit rate limit. Retrying in {retry_delay} seconds.")
                time.sleep(retry_delay)
                retry_delay *= 2
            except Exception as err:
                print(f"Error: {err}")
                break

        def _deanonymize(response):
            if output_format:
                for key, value in response.items():
                    response[key] = self.deanonymizer.deanonymize(value, mappings)
                return output_format.model_validate(response)
            else:
                return self.deanonymizer.deanonymize(response, mappings)

        deanonymized_responses = [_deanonymize(res) if self.anonymize else res
                                  for res in valid_responses]

        if n_completions == 1:
            # if generating a single completion, return it directly
            return deanonymized_responses[0]
        return deanonymized_responses

    def _model_structure_repr(self, model: Type[BaseModel]) -> str:
        fields = model.__annotations__
        return ', '.join(f'{key}: {value}' for key, value in fields.items())


    def _is_valid_json_for_model(self, text: str, model: Type[BaseModel]) -> bool: # noqa
        """
        Check if a text is valid JSON and if it respects the provided BaseModel. # noqa
        """
        model.model_config = ConfigDict(strict=True)

        try:
            parsed_data = json.loads(text)
            model(**parsed_data)
            return True
        except (json.JSONDecodeError, ValidationError):
            return False
